#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# === Dictionaries::AskEnglishWord
#
# Use this class to ask a random english word.
#
# There are two ways to use this class:
#
# - Pass in an optional argument to it, which we will use to query this
#   word from the database.
#
# - Pass nothing to it, in which case we load a random entry.
#
# Specific usage examples:
#
#   require 'dictionaries'
#   Dictionaries::AskEnglishWord.new
#   Dictionaries.ask_english_word
#
# =========================================================================== #
# require 'dictionaries/ask_english_word.rb'
# Dictionaries.path_to_the_english_file?
# =========================================================================== #
require 'dictionaries/class/ask_word_from_dictionary.rb'

module Dictionaries

class AskEnglishWord < AskWordFromDictionary

  # ========================================================================= #
  # === DATA
  # ========================================================================= #
  if defined?(DATA) == 'constant'
    if DATA.respond_to? :read
      _ = DATA.read
    else
      e 'DATA does not respond to :read'
    end
  else
    _ = ''
  end

  # ========================================================================= #
  # === DEFAULT_DELAY
  # ========================================================================= #
  if _.to_s.empty? # Next, define the default delay.
    DEFAULT_DELAY = 2.2 # How long to wait before asking a question.
  else
    _ = _.first if _.is_a? Array
    DEFAULT_DELAY = _.strip.to_f
  end

  # ========================================================================= #
  # === AskEnglishWord::THIS_FILE
  # ========================================================================= #
  THIS_FILE = THIS_FILE_HERE = __FILE__.to_s

  unless ENV['DICTIONARIES_FILE'] and File.exist?(ENV['DICTIONARIES_FILE'])
    Dictionaries.set_main_file(THIS_FILE) # Assign it at once.
  end

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      commandline_arguments = ARGV
    )
    super(commandline_arguments, self.class.main_file?)
  end

  # ========================================================================= #
  # === Dictionaries::AskEnglishWord.main_file?
  # ========================================================================= #
  def self.main_file?
    Dictionaries.main_file?
  end

  # ========================================================================= #
  # === Dictionaries::AskEnglishWord.dataset?
  #
  # Note that this method will load the dataset from the main file anew
  # whenever you invoke it.
  # ========================================================================= #
  def self.dataset?
    YAML.load_file(self.main_file?)
  end

  # ========================================================================= #
  # === Dictionaries::AskEnglishWord.n_entries?
  #
  # Feedback how many entries are registered.
  # ========================================================================= #
  def self.n_entries?
    _ = dataset?
    n_entries = _.size
    return n_entries
  end; self.instance_eval { alias n_questions? n_entries? } # === AskEnglishWord.n_questions?

  # ========================================================================= #
  # === Dictionaries::AskEnglishWord[]
  # ========================================================================= #
  def self.[](i)
    self.new(i)
  end

end

# =========================================================================== #
# === Dictionaries.ask_english_word
# =========================================================================== #
def self.ask_english_word
  AskEnglishWord.new
end

# =========================================================================== #
# === Dictionaries.path_to_the_english_file?
#
# This method may return a String, as a file path, such as
# "/home/Programs/Ruby/2.7.2/lib/ruby/site_ruby/2.7.0/dictionaries/yaml/english.yml".
# =========================================================================== #
def self.path_to_the_english_file?
  Dictionaries.main_file?
end

# =========================================================================== #
# === Dictionaries.n_entries?
# =========================================================================== #
def self.n_entries?
  ::Dictionaries::AskEnglishWord.n_entries?
end

end

if __FILE__ == $PROGRAM_NAME
  Dictionaries::AskEnglishWord.new(ARGV)
end # askeng
# This is the delay to use for this class.
__END__
1.8