#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# This is the base class for the various dictionaries.
#
# That way, we can ask specific words from a specific dictionary,
# such as english words or italian words and so forth.
# =========================================================================== #
# require 'dictionaries/class/ask_word_from_dictionary.rb'
# =========================================================================== #
require 'yaml'

# =========================================================================== #
# Require module-methods of this project next.
# =========================================================================== #
require 'dictionaries/toplevel_methods/main_file.rb'
# =========================================================================== #
# Next, load up dictionaries-related ruby code.
# =========================================================================== #
require 'dictionaries/constants.rb'
require 'dictionaries/class/colours.rb'
require 'dictionaries/class/set_search_for_this_word.rb'
require 'dictionaries/class/constants.rb'

module Dictionaries

class AskWordFromDictionary # === Dictionaries::AskWordFromDictionary

  begin
    require 'generate_shell_completion'
  rescue LoadError; end

  include Colours::E

  begin
    require 'opn'
  rescue LoadError; end

  begin
    require 'open'
  rescue LoadError; end

  begin
    require 'xorg_buffer/module'
    include XorgBuffer
  rescue LoadError; end

  begin
    require 'save_file/module'
  rescue LoadError; end

  # ========================================================================= #
  # === initialize
  # ========================================================================= #
  def initialize(
      optional_search_for_this_word = nil,
      use_this_file_for_the_dataset = :default_file,
      run_already                   = true
    )
    register_sigint
    reset
    case optional_search_for_this_word
    when :do_not_run_yet
      run_already = :do_not_run_yet
      optional_search_for_this_word = nil
    end
    # ======================================================================= #
    # === Next, set the main file to use
    # ======================================================================= #
    if use_this_file_for_the_dataset
      set_use_this_file(use_this_file_for_the_dataset)
      load_main_dataset
    end
    set_search_for_this_word(optional_search_for_this_word)
    if run_already.is_a? Hash
      if run_already.has_key? :delay
        set_delay(run_already.delete(:delay))
      end
      run_already = true
    end
    case run_already
    when :dont_run_yet, :do_not_run_yet
      run_already = false
    end
    run if run_already
  end

  # ========================================================================= #
  # === reset
  # ========================================================================= #
  def reset
    # ======================================================================= #
    # === @original_words
    # ======================================================================= #
    @original_words = {} # A Hash.
    # ======================================================================= #
    # === @words
    # ======================================================================= #
    @words = nil
    # ======================================================================= #
    # === @be_verbose
    # ======================================================================= #
    @be_verbose = true
    # ======================================================================= #
    # === @store_line_number_where
    # ======================================================================= #
    @store_line_number_where = STORE_LINE_NUMBER_HERE
    # ======================================================================= #
    # === @ask_the_question
    # ======================================================================= #
    @ask_the_question = true # Whether to ask the question or not.
    set_use_this_file # Set to nil initially.
    set_delay
  end

  # ========================================================================= #
  # === set_delay                                                 (delay tag)
  # ========================================================================= #
  def set_delay(i = DEFAULT_DELAY)
    @delay = i
  end

  # ========================================================================= #
  # === find_all_matches_for
  # ========================================================================= #
  def find_all_matches_for(i) # Must return an Array.
    selection = @words.select {|key,value| key.start_with? i }.keys
    return selection
  end

  # ========================================================================= #
  # === available_keys?
  # ========================================================================= #
  def available_keys?
    @words.keys if @words
  end; alias array available_keys? # === array

  # ========================================================================= #
  # === find_this_word_in_yaml_file
  #
  # This method will give us the correct line number.
  # ========================================================================= #
  def find_this_word_in_yaml_file(this_word)
    _ = File.readlines(
      @use_this_file, :encoding => USE_THIS_ENCODING
    )
    _ = _.index() {|x| x =~ /^#{this_word}/ }
    line_number = _.to_i + 1
    save_line_number(line_number)
    save_last_question_asked(this_word)
  end

  # ========================================================================= #
  # === show_every_word
  #
  # This will simply show every word.
  # ========================================================================= #
  def show_every_word
    @original_words.each_pair {|key, value|
      e "#{key.to_s.ljust(20)} -> #{value}"
    }
  end

  # ========================================================================= #
  # === return_random_word
  # ========================================================================= #
  def return_random_word
    available_keys?.sample
  end

  # ========================================================================= #
  # === generate_tab_completion                                     (tab tag)
  #
  # The first argument should be the name of the dictionary in question,
  # such as 'english' or 'italian'.
  #
  # To generate the tab completion, do this:
  #
  #   askeng TAB
  #
  # ========================================================================= #
  def generate_tab_completion(
      this_language = 'english'
    )
    require 'generate_shell_completion'
    dataset = @original_words.keys # cat $SCI/YAML/DICTIONARIES/english.yml
    # Next, determine where to store the completion stuff.
    generate_where =
      ENV['MY_OS'].to_s+'/LINUX/SHELL/SCRIPTS/'+this_language+'_completion.sh'
    _ = GenerateShellCompletion::GenerateShellCompletion.new(:do_not_run_yet)
    _.add_data(dataset)
    _.set_file(generate_where)
    _.set_name_of_function(this_language.to_sym)
    _.add_aliases %w( askeng ask_eng ) 
    _.generate
  end

  # ========================================================================= #
  # === reveal_the_answer
  #
  # Show the answer here.
  # ========================================================================= #
  def reveal_the_answer
    @result = @original_words[@search_for_this_word].to_s
    _ = @result
    begin
      _[0,1] = _[0,1].upcase
    rescue Exception => error
      opnn; e 'An exception happened at line '+simp(__LINE__.to_s)+' when '\
              'we tried to use .upcase.'
      opnn; e 'The file was '+sfile(__FILE__)
      opnn; e 'The difficult word in question was `'+simp(_)+'`.'
      opnn; e 'The encoding in use was '+simp(_.encoding.to_s)+'.'
      pp error
      begin
        _ = _.encode(USE_THIS_ENCODING)
        _[0,1] = _[0,1].upcase
      rescue Exception => error
        pp error
      end
    end
    _ = sanitize_answer(_)
    if _.include? '[]' # Chop off [] token, if it is part of the String.
      _ = _.force_encoding(USE_THIS_ENCODING) if _.encoding.to_s.include? 'UTF'
      _ = _.delete('[').delete(']').strip
    end
    e N+simportant(_)+'.'+N+N
  end; alias reveal_answer reveal_the_answer # === reveal_answer

  # ========================================================================= #
  # === register_sigint
  # ========================================================================= #
  def register_sigint
    Signal.trap('SIGINT') { exit }
  end

  # ========================================================================= #
  # === is_included?
  # ========================================================================= #
  def is_included?(i)
    @original_words.has_key?(i)
  end

  # ========================================================================= #
  # === feedback_help_options
  #
  # If you wish to trigger this, do:
  #
  #   askeng --help
  #
  # ========================================================================= #
  def feedback_help_options # help tag
    e
    e LPAD+sfancy('TOTAL')+'  # Feedback how many keys are registered, '+
      'also aliased to '+sfancy('STATS')+'.'
    e LPAD+sfancy('OPEN')+'   # Open '+sfile(File.basename(main_file?))+
      ' and '+sfile('ask_english_word.rb')+' in the main editor.'
    e LPAD+sfancy('SOLVED')+' # Set the last word as solved.'
    e LPAD+sfancy('URL')+'    # Show the URL for the dictionary.'
    e LPAD+sfancy('REPEAT')+' # Repeat the last question asked.'
    e LPAD+sfancy('DELAY?')+' # Feedback the delay we use.'
    e LPAD+sfancy('STATS')+'  # To show some statistics, especially '\
      'the amount of registered words.'
    e LPAD+sfancy('FILE?')+'  # Show the main .yml file in question.'
    e
    exit
  end

  # ========================================================================= #
  # === search_for?
  # ========================================================================= # 
  def search_for?
    @search_for_this_word
  end

  # ========================================================================= #
  # === words?
  # ========================================================================= #
  def words?
    @words
  end; alias dataset? words? # === dataset?

  # ========================================================================= #
  # === be_verbose?
  # ========================================================================= #
  def be_verbose?
    @be_verbose
  end

  # ========================================================================= #
  # === result?
  # ========================================================================= #
  def result?
    @result
  end; alias result result? # === result

  # ========================================================================= #
  # === words?
  # ========================================================================= #
  def words?
    @words
  end

  # ========================================================================= #
  # === sanitize_dataset
  # ========================================================================= #
  def sanitize_dataset # We have to ignore all entries that contain []. We do this only for @words.
    words?.reject! {|key, value|
      value.include? '[]'
    }
  end

  # ========================================================================= #
  # === on_roebe?
  # ========================================================================= #
  def on_roebe?
    ENV['IS_ROEBE'].to_s == '1'
  end

  # ========================================================================= #
  # === do_ask_the_question
  # ========================================================================= #
  def do_ask_the_question # This method will ask the english question.
    if search_for?.empty?
      set_search_for_this_word 'RANDOM'
    end
    _ = search_for?
    if _ == 'rand'
      _ = (@search_for_this_word = return_random_word)
    end
    unless is_included? _
      set_xorg_buffer _
      opnn; e 'Sorry, the word `'+sfancy(_)+'` is not included.'
      opnn; e 'We will thus try to open our english dictionary in our editor'
      opnn; e 'and lookup the word in our browser.'
      open_main_yaml_file_in_main_editor # Added ~ 22.02.2013
      consider_opening_browser_page(_)
    else # Ok, we now know that it is included. Thus, continue.
      if be_verbose?
        e "#{rev}What does the word `#{sfancy(_)}` mean? #{show_delay_string}"
        find_this_word_in_yaml_file(_)
        new_buffer = build_url_www_string(_)
        set_xorg_buffer(new_buffer)
        sleep delay?
        reveal_the_answer
      end
    end
  end

  # ========================================================================= #
  # === rev
  # ========================================================================= #
  def rev
    Colours.rev
  end

  # ========================================================================= #
  # === set_xorg_buffer
  # ========================================================================= #
  def set_xorg_buffer(i)
    XorgBuffer[i] if Object.const_defined? :XorgBuffer
  end

  # ========================================================================= #
  # === store_last_question_asked_where?
  #
  # This method will try to automatically find out where we will store
  # the last question asked.
  # ========================================================================= #
  def store_last_question_asked_where?
    name = Dictionaries.return_name_from_compound(main_file?)
    path = DEPOT_INFORMATION_DIR+'last_'+name+'_question_asked'
    return path
  end

  # ========================================================================= #
  # === consider_opening_browser_page
  # ========================================================================= #
  def consider_opening_browser_page(i = '')
    url = build_url_www_string(i)
    opnn; e 'Consider visiting '+brown(url)+' if you need a translation.'
    open_in_browser url
  end

  # ========================================================================= #
  # === build_url_www_string
  # ========================================================================= #
  def build_url_www_string(i)
    _ = ''.dup
    # ======================================================================= #
    # We need to differ between the english dictionary and the italian
    # dictionary.
    # ======================================================================= #
    language = Dictionaries.return_name_from_compound(main_file?)
    if language == 'english'
      _ << '"' unless _.end_with? '"'
    end
    case language # case tag
    when 'italian'
      _ << URL_FOR_ITALIAN_DICTIONARY
    when 'english' # This is the default.
      _ << URL_FOR_DICT_LEO+'ende/?lp=ende&lang=de&searchLoc=0&cmpType=relaxed&sectHdr=on&spellToler=&search='
    end
    _ << i.to_s
    if language == 'english'
      _ << '"' unless _.end_with? '"' # Append '"' here.
    end
    _
  end

  # ========================================================================= #
  # === show_help
  #
  # Show the help options.
  # ========================================================================= #
  def show_help
  end

  # ========================================================================= #
  # === load_dataset
  #
  # Here we will define the important instance variable @original_words.
  # ========================================================================= #
  def load_dataset(
      of_this_file = @use_this_file
    ) # called via reset()
    of_this_file = of_this_file.to_s # We must work with a String.
    if File.exist? of_this_file
      dataset = YAML.load_file(of_this_file)
      @original_words = dataset
      # ^^^ This is the file english.yml
    else
      e 'Did not find a file at '+of_this_file+'.'
      e 'Thus, the @original_words will be empty.'
      e 'If you wish to change this, modify the constant ENGLISH_WORDS'
      e 'to point at a yaml-dataset (we require a hash, thus key: value pairs)'
      @original_words = {}
    end
    @words = @original_words.dup
    sanitize_dataset # Get rid of lines that have [] as entry.
  end; alias load_main_dataset load_dataset # === load_main_dataset

  # ========================================================================= #
  # === set_use_this_file
  #
  # This sets the main dictionary-file to use, the file path to it.
  # ========================================================================= #
  def set_use_this_file(i = :default)
    case i
    when :english
      i = ENGLISH_WORDS
    when :default_file, :default
      i = main_file?
    end
    @use_this_file = i
  end

  # ========================================================================= #
  # === load_the_english_file
  # ========================================================================= #
  def load_the_english_file
    set_use_this_file(:english)
    load_main_dataset
  end

  # ========================================================================= #
  # === use_this_file?
  # ========================================================================= #
  def use_this_file?
    @use_this_file
  end

  # ========================================================================= #
  # === opnn
  # ========================================================================= #
  def opnn
    Opn.opn(namespace: NAMESPACE)
  end

  # ========================================================================= #
  # === feedback_statistics
  #
  # Feedback the amount of questions registered. In the future, perhaps
  # we may want to have more statistic-options, so this may be expanded
  # at a later time.
  #
  # To trigger this, do:
  #   askeng --stats?
  # ========================================================================= #
  def feedback_statistics
    feedback_total_amount_of_keys(:do_not_exit)
  end

  # ========================================================================= #
  # === feedback_url
  # ========================================================================= #
  def feedback_url
    opnn; e "  -> #{URL_FOR_DICT_LEO}"
    exit
  end

  # ========================================================================= #
  # === open_yaml_file_in_main_editor
  # ========================================================================= #
  def open_yaml_file_in_main_editor(
      this_file = main_file?
    )
    if on_roebe? # Rewrite for my own system at home.
      this_file = File.basename(this_file)
      this_file = MY_DICTIONARIES+this_file
    end
    _= this_file
    notify_the_user_that_we_will_next_open_this_file(_)
    system "bluefish #{_}"
  end; alias open_main_yaml_file_in_main_editor open_yaml_file_in_main_editor # === open_main_yaml_file_in_main_editor

  # ========================================================================= #
  # === notify_the_user_that_we_will_next_open_this_file
  # ========================================================================= #
  def notify_the_user_that_we_will_next_open_this_file(this_file)
    opnn; e 'Opening file `'+sfile(this_file)+'` next.'
  end

  # ========================================================================= #
  # === feedback_total_amount_of_keys
  # ========================================================================= #
  def feedback_total_amount_of_keys(shall_we_exit = true) # KEYS
    case shall_we_exit
    when :do_not_exit
      shall_we_exit = false
    end
    _ = @original_words.size - @words.size
    opnn; e swarn(@original_words.size.to_s)+' english words in total.'
    rounded = sprintf('%0.2f', 100.0 - (@words.size * 100.0 / @original_words.size) )
    opnn; e swarn(@words.size.to_s)+' unanswered english words, '\
            'thus '+sfancy(_.to_s)+' are already answered. ('+
            simp(rounded.to_s+'%')+')'
    opnn; e 'The yaml file that we we will use is:'
    opnn; e '  '+sfile(main_file?)
    exit if shall_we_exit
  end

  # ========================================================================= #
  # === open_this_file_here
  # ========================================================================= #
  def open_this_file_here(be_verbose = false)
    _= Dictionaries.main_file?
    if be_verbose
      opnn; notify_the_user_that_we_will_next_open_this_file(_)
    end
    system "bluefish #{_}"
  end

  # ========================================================================= #
  # === sanitize_answer
  #
  # This may also re-encode the given String object.
  # ========================================================================= #
  def sanitize_answer(i)
    begin
      unless i.encoding.to_s.include? USE_THIS_ENCODING
        i = i.force_encoding(USE_THIS_ENCODING)
      end
      i.gsub!(/ \[\]/,'') if i.include? '[]'
    rescue Exception => error
      opnn; pp error # Propagate the error here to the user.
    end
    return i
  end

  # ========================================================================= #
  # === save_line_number
  # ========================================================================= #
  def save_line_number(what) # save tag
    begin
      save_file(what, @store_line_number_where)
    rescue => error
      pp error
      pp error.class
    end
  end

  # ========================================================================= #
  # === set_last_word_solved
  # ========================================================================= #
  def set_last_word_solved # Call this only when you want to solve an english word.
    begin
      require 'roebe/classes/append_to_line.rb'
      _ = Roebe::AppendToLine.new(:english, false) # bl $RUBY_TOOLS/append_to_line.rb
      _.set_line File.read(STORE_LINE_NUMBER_HERE)
      _.run
    rescue LoadError; end  
    exit
  end

  # ========================================================================= #
  # === save_last_question_asked
  # ========================================================================= #
  def save_last_question_asked(what) # Save the last question asked.
    into = store_last_question_asked_where?
    begin
      save_file(what, into)
    rescue Errno::ENOENT
      opnn; e "Can not store into `#{sfile(into)}`."
    end
  end

  # ========================================================================= #
  # === save_file
  # ========================================================================= #
  def save_file(what, into)
    SaveFile.write_what_into(what, into)
  end

  # ========================================================================= #
  # === delay?
  # ========================================================================= #
  def delay?
    @delay
  end

  # ========================================================================= #
  # === feedback_delay
  # ========================================================================= #
  def feedback_delay
    opnn; e "The delay we use is set to: #{sfancy(@delay.to_s)}"
    exit
  end

  # ========================================================================= #
  # === translate
  # ========================================================================= #
  def translate(i)
    return @original_words[i].to_s
  end

  # ========================================================================= #
  # === show_delay_string
  # ========================================================================= #
  def show_delay_string
    result = '(Delay is at '+simp(@delay.to_s)+' seconds)'
    result = '' if @delay.to_i == 0
    return result
  end

  # ========================================================================= #
  # === feedback_yaml_file
  # ========================================================================= #
  def feedback_yaml_file
    efile main_file?
  end

  # ========================================================================= #
  # === main_file?
  # ========================================================================= #
  def main_file?
    Dictionaries.main_file?
  end; alias currently_selected_file? main_file? # === currently_selected_file?

  # ========================================================================= #
  # === run                                                         (run tag)
  # ========================================================================= #
  def run
    load_dataset
    do_ask_the_question if @ask_the_question
  end

end

# =========================================================================== #
# === Dictionaries.translate
# =========================================================================== #
def self.translate(this_word = nil)
  _ = Dictionaries::AskWordFromDictionary.new(this_word, :default_file, :dont_run_yet)
  _.translate(this_word)
end

# =========================================================================== #
# === Dictionaries.new
# =========================================================================== #
def self.new(i = nil)
  Dictionaries::AskWordFromDictionary.new(i)
end

end

if __FILE__ == $PROGRAM_NAME
  pp Dictionaries::AskWordFromDictionary.new
  pp Dictionaries::AskWordFromDictionary.class
  pp Dictionaries::DICTIONARIES_DIR
  pp Dictionaries::ENGLISH_WORDS
end