#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'dictionaries/toplevel_methods/misc.rb'
# =========================================================================== #
module Dictionaries

  require 'yaml'
  require 'dictionaries/constants.rb'
  require 'dictionaries/toplevel_methods/main_file.rb'
  require 'dictionaries/toplevel_methods/has_key.rb'

  # ========================================================================= #
  # === Dictionaries.generate_pdf_file
  #
  # This method can be used to generate a .pdf file.
  # ========================================================================= #
  def self.generate_pdf_file(
      use_this_font = 'Courier', # Helvetica-Bold' #  #'Times-Roman'
      use_this_font_size = 12
    )
    require 'prawn'
    ::Prawn::Fonts::AFM.hide_m17n_warning = true
    right_arrow = ' -> '.dup.encode(
                    'Windows-1252', invalid: :replace, undef: :replace, replace: ''
                  )
    dataset = YAML.load_file(Dictionaries.file_english)
    into = File.absolute_path('english_to_german_dictionary.pdf')
    Prawn::Document.generate(into) {
      font use_this_font
      font_size use_this_font_size
      text(
        "English to German dictionary: "\
        "#{dataset.keys.size.to_s} translated words\n\n"
      )
      # ===================================================================== #
      # Iterate over our exam dataset next.
      # ===================================================================== #
      dataset.each_pair {|key, value|
        indent(8) {
          key   = key.dup if key.frozen?
          value = value.dup if value.frozen?
          result = key.encode(
              'Windows-1252', invalid: :replace, undef: :replace, replace: ''
            )+
            right_arrow+
            value.encode(
              'Windows-1252', invalid: :replace, undef: :replace, replace: ''
            )
          text(result, {size: 10})
        }
      }
    }
    e 'Stored into `'+into+'`.'
  end

  # ========================================================================= #
  # === Dictionaries[]
  #
  # Currently this method will always reload the main file. In the future
  # we may change this approach, but for now this has to suffice (May 2019).
  #
  # The method will return either the translated string - or nil. Nil
  # indicates that the main Hash does not include that key; in other words,
  # that the word is not registered.
  # ========================================================================= #
  def self.[](i = ARGV)
    i = i.join.strip if i.is_a? Array
    this_file = Dictionaries.main_file?
    if this_file.nil?
      Dictionaries.set_main_file(:default) # Must initialize it in this case.
      this_file = Dictionaries.main_file?
    end
    if this_file
      dataset = YAML.load_file(this_file)
      dataset[i] # Return the result here.
    else
      nil
    end
  end

  # ========================================================================= #
  # === Dictionaries.return_array_of_translated_words_from_online_leo
  #
  # This method can be used to query the online dictionary from leo.
  #
  # It will then return an Array of translations.
  #
  # Note that this currently only works for the translation from english 
  # to german, not the other way around, even though that is not too 
  # hard to implement either.
  # ========================================================================= #
  def self.return_array_of_translated_words_from_online_leo(this_word)
    if this_word.is_a? Array
      this_word = this_word.join(' ').strip
    end
    require 'open-uri'
    remote_url = "https://dict.leo.org/german-english/#{this_word}"
    remote_dataset = URI.open(remote_url).read
    # ======================================================================= #
    # See: https://rubular.com/r/OOXwAc6PVqjU5Q
    # ======================================================================= #
    use_this_regex =
      /<words><word>([a-zA-Z\s]+)\<\/word><\/words>/
    scanned_results = remote_dataset.scan(use_this_regex).flatten
    # ======================================================================= #
    # This result may look like so:
    #   [["cat"], ["die Katze"]
    # We have to sanitize it still.
    # ======================================================================= #
    scanned_results.reject! {|entry| entry.start_with?(this_word) }
    return scanned_results # Note that the full Array is returned.
  end

  # ========================================================================= #
  # === Dictionaries.return_unique_words_from_this_file
  #
  # This method will read in the words from an existing local file.
  #
  # Some files may be huge, though, and then this method becomes
  # quite useless, so we really should discard a lot of data after a
  # certain threshold, to keep memory usage low.
  # ========================================================================= #
  def self.return_unique_words_from_this_file(
      this_file,
      optional_arguments = :english,
      remove_html_tags   = false
    )
    case remove_html_tags
    # ======================================================================= #
    # === :remove_HTML_tags
    # ======================================================================= #
    when :remove_HTML_tags
      remove_html_tags = true
    end
    dataset = nil
    if this_file.is_a? Array
      this_file = this_file.join(' ').strip
    end
    if this_file and File.file?(this_file)
      dataset = File.read(this_file)
      if remove_html_tags
        require 'cyberweb/toplevel_methods/misc.rb'
        dataset = ::Cyberweb.remove_html(dataset)
      end
    end
    result = [] # ← This variable will store the unique words found in the file.
    case optional_arguments
    # ======================================================================= #
    # === :english
    # ======================================================================= #
    when :english,
         :default
      # ===================================================================== #
      # Query unknown words in english. We will compare the words to the
      # available variants.
      # ===================================================================== #
      result = dataset.scan(/\w+/).map {|entry| entry.downcase } # ← Obtain all words here.
      # ===================================================================== #
      # Next, reject those that are registered:
      # ===================================================================== #
      result.reject! {|this_word|
        ::Dictionaries.has_key?(this_word)
      }
    end if dataset
    return result.uniq.sort
  end

end

if __FILE__ == $PROGRAM_NAME
  alias e puts
  #puts Dictionaries[ARGV]
  e 'Possible words for cat:'
  pp Dictionaries.return_array_of_translated_words_from_online_leo('cat')
  # ========================================================================= #
  # The next chunk of code is to test reading unique words from a given
  # local file.
  # ========================================================================= #
  puts '-'*80
  _ = Dictionaries.return_unique_words_from_this_file('/Depot/j/foobar.md')
  pp _
  pp _.size
end # rb misc.rb cat